/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "registers.h"

/**
 * FID (Function ID):
 * - Bits 15..12 (FID Class)
 * - Bits 11..6 (FID Subclass)
 * - Bits 5..0 (Function Module)
 */

#define FidGetFunctionModule(fid_)  ((fid_) & 0x3F)

/*..........................................................................*/
bool Fid_HasButton(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_STANDARD_R100:
            switch (FidGetFunctionModule(fid) & FM_BUTTON_CFG_MASK) {
            case FM_BUTTON_CFG_1X:
            case FM_BUTTON_CFG_2X:
            case FM_BELL_BUTTON_4WAY:
                return TRUE;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_1WAY_R105:
            return ((FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK)? TRUE : FALSE);
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON;
        default:
            break;
        }
        break;
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
            return ((FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK)? TRUE : FALSE);
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON;
        }
        break;
    case FID_CLASS_SECURITY:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_SECURITY_ALARM:
            switch (FidGetFunctionModule(fid) & FM_ALARM_BUTTON_MASK) {
            case FM_ALARM_BUTTON_ALARM:
            case FM_ALARM_BUTTON_PANIC:
            case FM_ALARM_BUTTON_ALARM_PANIC:
                return TRUE;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_1WAY_R105:
            return ((FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK)? TRUE : FALSE);
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON;
        default:
            break;
        }
        break;
    case FID_CLASS_ACCESS:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_ACCESS_BELL:
            switch (FidGetFunctionModule(fid) & FM_BUTTON_CFG_MASK) {
            case FM_BUTTON_CFG_1X:
            case FM_BUTTON_CFG_2X:
            case FM_BELL_BUTTON_4WAY:
                return TRUE;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_ACCESS_LEAVING:
            switch (FidGetFunctionModule(fid) & FM_LEAVE_BUTTON_MASK) {
            case FM_LEAVE_BUTTON_LEAVE:
            case FM_LEAVE_BUTTON_COME:
            case FM_LEAVE_BUTTON_LEAVE_COME:
                return TRUE;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_ACCESS_LOCKING:
            break;
        case FID_SUBCLASS_1WAY_R105:
            return ((FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK)? TRUE : FALSE);
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return FALSE;
}

/*..........................................................................*/
uint8_t Fid_NumberOfsButtons(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_STANDARD_R100:
            switch (FidGetFunctionModule(fid) & FM_BUTTON_CFG_MASK) {
            case FM_BUTTON_CFG_1X:
                return 1;
            case FM_BUTTON_CFG_2X:
                return 2;
            case FM_BELL_BUTTON_4WAY:
                return 4;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_1WAY_R105:
            switch (FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK) {
            case FM_R105_BUTTON_CFG_1X:
                return 1;
            case FM_R105_BUTTON_CFG_2X:
                return 2;
            case FM_R105_BUTTON_CFG_4X:
                return 4;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_PRE_R200:
            return ((FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON) ? 1 : 0;
        default:
            break;
        }
        break;
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
            switch (FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK) {
            case FM_R105_BUTTON_CFG_1X:
                return 1;
            case FM_R105_BUTTON_CFG_2X:
                return 2;
            case FM_R105_BUTTON_CFG_4X:
                return 4;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_PRE_R200:
            return ((FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON) ? 1 : 0;
        }
        break;
    case FID_CLASS_SECURITY:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_SECURITY_ALARM:
            switch (FidGetFunctionModule(fid) & FM_ALARM_BUTTON_MASK) {
            case FM_ALARM_BUTTON_ALARM:
            case FM_ALARM_BUTTON_PANIC:
            case FM_ALARM_BUTTON_ALARM_PANIC:
                return 1;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_1WAY_R105:
            switch (FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK) {
            case FM_R105_BUTTON_CFG_1X:
                return 1;
            case FM_R105_BUTTON_CFG_2X:
                return 2;
            case FM_R105_BUTTON_CFG_4X:
                return 4;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_PRE_R200:
            return ((FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON) ? 1 : 0;
        default:
            break;
        }
        break;
    case FID_CLASS_ACCESS:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_ACCESS_BELL:
            switch (FidGetFunctionModule(fid) & FM_BUTTON_CFG_MASK) {
            case FM_BUTTON_CFG_1X:
            case FM_BUTTON_CFG_2X:
            case FM_BELL_BUTTON_4WAY:
                return 1;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_ACCESS_LEAVING:
            switch (FidGetFunctionModule(fid) & FM_LEAVE_BUTTON_MASK) {
            case FM_LEAVE_BUTTON_LEAVE:
            case FM_LEAVE_BUTTON_COME:
            case FM_LEAVE_BUTTON_LEAVE_COME:
                return 1;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_ACCESS_LOCKING:
            break;
        case FID_SUBCLASS_1WAY_R105:
            switch (FidGetFunctionModule(fid) & FM_R105_BUTTON_MASK) {
            case FM_R105_BUTTON_CFG_1X:
                return 1;
            case FM_R105_BUTTON_CFG_2X:
                return 2;
            case FM_R105_BUTTON_CFG_4X:
                return 4;
            default:
                break;
            }
            break;
        case FID_SUBCLASS_PRE_R200:
            return ((FidGetFunctionModule(fid) & FM_PRE_R200_BUTTON) == FM_PRE_R200_BUTTON) ? 1 : 0;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return 0;
}

/*..........................................................................*/
bool Fid_HasVariableButtonIdGroup(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_STANDARD_R100:
            return (FidGetFunctionModule(fid) & FM_VAR_BTN_GRP_MASK) == FM_VAR_BTN_GRP_CFG;
        case FID_SUBCLASS_1WAY_R105:
        case FID_SUBCLASS_SCENE_DEVICE:
        case FID_SUBCLASS_PRE_R200:
            return TRUE;
        default:
            break;
        }
        break;
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
        case FID_SUBCLASS_SCENE_DEVICE:
        case FID_SUBCLASS_PRE_R200:
            return TRUE;
        }
        break;
    case FID_CLASS_SECURITY:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_SECURITY_ALARM:
            return FALSE;
        case FID_SUBCLASS_1WAY_R105:
        case FID_SUBCLASS_SCENE_DEVICE:
        case FID_SUBCLASS_PRE_R200:
            return TRUE;
        default:
            break;
        }
        break;
    case FID_CLASS_ACCESS:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_ACCESS_BELL:
        case FID_SUBCLASS_ACCESS_LEAVING:
        case FID_SUBCLASS_ACCESS_LOCKING:
            return FALSE;
        case FID_SUBCLASS_1WAY_R105:
        case FID_SUBCLASS_SCENE_DEVICE:
        case FID_SUBCLASS_PRE_R200:
            return TRUE;
        default:
            break;
        }
        break;
    case FID_CLASS_OVERALL: {
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_COLOR_SCENE_DEVICE:
            return TRUE;
        default:
            break;
        }
        break;
    }
    default:
        break;
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_HasOutput(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_STANDARD_R100:
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_OUTPUT_PRESENT) == FM_OUTPUT_PRESENT;
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_OUTPUT) == FM_PRE_R200_OUTPUT;
        default:
            break;
        }
        break;
    case FID_CLASS_JOKER:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_OUTPUT_PRESENT) == FM_OUTPUT_PRESENT;
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_OUTPUT) == FM_PRE_R200_OUTPUT;
        default:
            break;
        }
        break;
    case FID_CLASS_SECURITY:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_SECURITY_ALARM:
        /*case FID_SUBCLASS_STANDARD_R100:*/
            /* #10322 legacy panic/alarm terminal blocks have the output bit wrong^winverted */
            return (FidGetFunctionModule(fid) & FM_OUTPUT_PRESENT) == 0;
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_OUTPUT_PRESENT) == FM_OUTPUT_PRESENT;
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_OUTPUT) == FM_PRE_R200_OUTPUT;
        default:
            break;
        }
        break;
    case FID_CLASS_ACCESS:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_ACCESS_BELL:
        case FID_SUBCLASS_ACCESS_LEAVING:
        case FID_SUBCLASS_ACCESS_LOCKING:
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_OUTPUT_PRESENT) == FM_OUTPUT_PRESENT;
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_OUTPUT) == FM_PRE_R200_OUTPUT;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_IsSecurityAlarmButton(uint16_t fid, uint8_t key)
{
    if ((FidGetClass(fid) == FID_CLASS_SECURITY) && (FidGetSubclass(fid) == FID_SUBCLASS_SECURITY_ALARM)) {
        uint8_t functionModule = FidGetFunctionModule(fid);
        switch (functionModule & FM_ALARM_BUTTON_MASK) {
        case FM_ALARM_BUTTON_ALARM: {
            return TRUE;
        }
        case FM_ALARM_BUTTON_ALARM_PANIC: {
            if (key == 0) {
                return TRUE;
            }
            break;
        }
        }
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_IsSecurityPanicButton(uint16_t fid, uint8_t key)
{
    if ((FidGetClass(fid) == FID_CLASS_SECURITY) && (FidGetSubclass(fid) == FID_SUBCLASS_SECURITY_ALARM)) {
        uint8_t functionModule = FidGetFunctionModule(fid);
        switch (functionModule & FM_ALARM_BUTTON_MASK) {
        case FM_ALARM_BUTTON_PANIC: {
            return TRUE;
        }
        case FM_ALARM_BUTTON_ALARM_PANIC: {
            if (key == 1) {
                return TRUE;
            }
            break;
        }
        }
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_IsAccessLeaveButton(uint16_t fid, uint8_t key)
{
    if ((FidGetClass(fid) == FID_CLASS_ACCESS) && (FidGetSubclass(fid) == FID_SUBCLASS_ACCESS_LEAVING)) {
        uint8_t functionModule = FidGetFunctionModule(fid);
        bool leaveButton = FALSE;
        switch (functionModule & FM_LEAVE_BUTTON_MASK) {
        case FM_LEAVE_BUTTON_LEAVE: {
            leaveButton = TRUE;
            break;
        }
        case FM_LEAVE_BUTTON_LEAVE_COME: {
            if (key == 0) {
                leaveButton = TRUE;
            }
            break;
        }
        }
        return leaveButton;
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_IsJocker(uint16_t fid)
{
    return (FidGetClass(fid) == FID_CLASS_JOKER);
}

/*..........................................................................*/
bool Fid_IsGroupMatching(uint16_t fid, uint8_t group)
{
    return (FidGetClass(fid) == group);
}

/*..........................................................................*/
bool Fid_IsdSLinkDevice(uint16_t fid)
{
    return ((FidGetClass(fid) == FID_CLASS_CLIMATE) &&
            (FidGetSubclass(fid) == FID_SUBCLASS_CUSTOM_DEVICE) &&
            (FidGetFunctionModule(fid) == FM_CLIMATE_CUSTOM_BS2));
}

/*..........................................................................*/
bool Fid_HasBinaryInputs(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
    case FID_CLASS_SECURITY:
    case FID_CLASS_ACCESS:
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_ADDITONAL_HW) == FM_ADDITONAL_HW;
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_BINARY_INPUT) == FM_PRE_R200_BINARY_INPUT;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return FALSE;
}

/*..........................................................................*/
bool Fid_HasSensors(uint16_t fid, uint32_t serial_number)
{
    int number_buttons = Fid_NumberOfsButtons(fid);
    if ((number_buttons > 1) && (serial_number % number_buttons != 0)) {
        return FALSE;
    }
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
    case FID_CLASS_SECURITY:
    case FID_CLASS_ACCESS:
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_SENSOR) == FM_PRE_R200_SENSOR;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return TRUE;
}

/*..........................................................................*/
bool Fid_HasOutputChannels(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
    case FID_CLASS_SECURITY:
    case FID_CLASS_ACCESS:
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_PRE_R200:
            return (FidGetFunctionModule(fid) & FM_PRE_R200_OUTPUT) == FM_PRE_R200_OUTPUT;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return FALSE;
}

/*..........................................................................*/
bool Ver_HasPbchannelPbgroupReg(uint16_t ver)
{
    return (ver >= 0x0358);
}

/*..........................................................................*/
bool Fid_HasVariableStandardGroup(uint16_t fid)
{
    switch (FidGetClass(fid)) {
    case FID_CLASS_LIGHT:
    case FID_CLASS_SHADE:
    case FID_CLASS_CLIMATE:
    case FID_CLASS_AUDIO:
    case FID_CLASS_VIDEO:
    case FID_CLASS_SECURITY:
    case FID_CLASS_ACCESS:
    case FID_CLASS_JOKER:
    case FID_CLASS_WHITE:
        switch (FidGetSubclass(fid)) {
        case FID_SUBCLASS_1WAY_R105:
            return (FidGetFunctionModule(fid) & FM_VARIABLE_STD_GROUP) == FM_VARIABLE_STD_GROUP;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return FALSE;
}
